# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT


import os
from typing import Optional, Tuple, Union

import abutils
from abutils import Sequence
from natsort import natsorted

from .antibody import Antibody

__all__ = [
    "get_germline_database_path",
    "get_germline",
    # "realign_germline",
    # "reassign_dgene",
    # "process_vgene_alignment",
    # "process_jgene_alignment",
    # "process_dgene_alignment",
]


# ------------------------------
#      GERMLINE GENES
# ------------------------------


def get_germline_database_path(germdb_name: str, receptor: str = "bcr") -> str:
    """
    Get the path to a germline database. The addon directory (which contains user-built
    databases) is checked first, and if a user-built database is not found, the built-in
    database is used if present.

    Parameters
    ----------
    germdb_name : str
        The name of the germline database. Typically the species (like "human"), but can also be a
        custom database (like "humouse") that merges multiple species or is in some other way not
        species-specific.

        .. note::
            All germline database names are lowercase, and identifying germline databases
            by name is case-insensitive.

    receptor : str, default: "bcr"
        The receptor type. Options are "bcr" and "tcr".

    Returns
    -------
    str
        Path to the germline database directory. In the germline directory will be the
        following subdirectories:

          - ``ungapped``: contains ungapped germline sequences in FASTA format (``"v.fasta"``, ``"d.fasta"``, etc.)
          - ``gapped``: contains IMGT-gapped germline sequences in FASTA format (``"v.fasta"``, ``"d.fasta"``, etc.)
          - ``mmseqs``: contains pre-compiled MMseqs2 databases

    Raises
    ------
    FileNotFoundError
        If the germline database is not found.

    ValueError
        If the receptor type is not one of "bcr" or "tcr".

    """
    germdb_name = germdb_name.lower()
    receptor = receptor.lower()
    if receptor not in ["bcr", "tcr"]:
        raise ValueError(f"Receptor type {receptor} not supported")
    # check the addon directory first
    addon_dir = os.path.expanduser(f"~/.abstar/germline_dbs/{receptor}")
    if os.path.isdir(addon_dir):
        if germdb_name.lower() in [os.path.basename(d[0]) for d in os.walk(addon_dir)]:
            return os.path.join(addon_dir, f"{receptor}/{germdb_name}")
    # if a user-generated DB isn't found, use the built-in DB
    core_dir = os.path.dirname(os.path.abspath(__file__))  # "abstar/core" directory
    abstar_dir = os.path.dirname(core_dir)  # "abstar" directory
    germdb_path = os.path.join(abstar_dir, f"germline_dbs/{receptor}/{germdb_name}")
    if not os.path.exists(germdb_path):
        raise FileNotFoundError(
            f"Germline database {germdb_name} for receptor {receptor} not found"
        )
    return germdb_path


def get_germline(
    germline_gene: str,
    germdb_name: str,
    receptor: str = "bcr",
    imgt_gapped: bool = False,
    exact_match: bool = False,
    force_constant: bool = False,
    truncate_species: bool = True,
) -> Union[list, Sequence]:
    """
    Get the germline sequence for a given germline gene.

    Parameters
    ----------
    germline_gene : str
        The germline gene to get the sequence for. Can be a full germline gene name including allele ('IGHV1-2*02'),
        or a truncated version corresponding to just the gene ('IGHV1-2') or the family ('IGHV1').

        .. warning::
            Querying with gene names that are a substrings of another gene name (for example, 'IGHV1-2' is a substring
            of 'IGHV1-24') will match and return both 'IGHV1-2' and 'IGHV1-24' genes. To prevent this, you can add an
            asterisk to the end of the query gene ('IGHV1-2*') which will cause the query to match only alleles of the
            'IGHV1-2' gene.

        .. note::
            Germline genes are organized into separate files by segment ("v", "d", "j", and "c"). Segments are parsed as
            the fourth character of the germline gene name. For example, 'IGHV1-2*02' is segment 'v'.

            Isotype queries (ex. IGHG1*01 or IGHM*02) are all inferred as segment type 'c'.

    germdb_name : str
        The name of the germline database.

    receptor : str, default: "bcr"
        The receptor type. Options are "bcr" and "tcr".

    imgt_gapped : bool, default: False
        Whether to use the IMGT gapped or ungapped germline sequences.

    force_constant : bool, default: False
        Whether to force the query to match a constant germline gene. This is necessary because both IgD and diversity (D)
        gene names are formatted as IGHD, making it ambiguous whether the query is for IgD or diversity. By default,
        a supplied germline name of the format IGHD is assumed to be diversity.

    truncate_species : bool, default: True
        Whether to truncate the species name from the germline gene name (if it is present). Truncation
        happens before matching, so if `truncate_species` is ``False`` and `exact_match` is ``True``, the query
        must contain the species name if the germline gene name is to be found in the database.

    Returns
    -------
    list or Sequence
        If multiple matches are found (for example, multiple alleles of the query 'IGHV1-2'), a list of Sequence objects is returned.
        If only one match is found (for example, when querying 'IGHV1-2*02'), a single Sequence object is returned.

    Raises
    ------
    FileNotFoundError
        If the germline database file is not found.

    ValueError
        If the germline gene segment is not one of 'v', 'd', 'j', or 'c'. The germline gene segment is inferred as the
        fourth character of the germline gene name. For example, 'IGHV1-2*02' is segment 'v'. Isotype queries (ex.
        IGHG1*01 or IGHM*02) are all inferred as segment type 'c'.

    ValueError
        If no germline genes matching the query are found in the database.

    """
    germdb_path = get_germline_database_path(germdb_name, receptor)
    # parse the segment type
    segment = germline_gene[3].lower()
    if segment in ["a", "g", "e", "m"] or force_constant:
        segment = "c"
    if segment not in ["v", "d", "j", "c"]:
        raise ValueError(
            f"The segment type must be one of 'v', 'd', 'j', or an isotype ('a', 'd', 'e', 'g', 'm'). Your input was {germline_gene}, which corresponds to segment type {segment}."
        )
    # decide on gapped vs ungapped
    if imgt_gapped:
        germdb_file = os.path.join(germdb_path, f"imgt_gapped/{segment}.fasta")
    else:
        germdb_file = os.path.join(germdb_path, f"ungapped/{segment}.fasta")
    if not os.path.exists(germdb_file):
        raise FileNotFoundError(
            f"Germline database {germdb_name} for receptor {receptor} not found"
        )
    # retrieve germline sequences
    all_germs = abutils.io.read_fasta(germdb_file)
    if truncate_species:
        for g in all_germs:
            g.id = g.id.split("__")[0]
    # do the matching
    if exact_match:
        germs = [g for g in all_germs if germline_gene == g.id]
    else:
        germs = [g for g in all_germs if germline_gene in g.id]
    if not germs:
        raise ValueError(
            f"No substring matches to {germline_gene} were found in {germdb_file}"
        )
    if len(germs) == 1:
        return germs[0]
    else:
        return natsorted(germs, key=lambda x: x.id)


# ------------------------------
#     GERMLINE ALILGNMENT
# ------------------------------


def realign_germline(
    sequence: str,
    germline_name: str,
    germdb_name: str,
    imgt_gapped: bool = False,
    semiglobal_aln_params: Optional[dict] = None,
    local_aln_params: Optional[dict] = None,
    skip_semiglobal: bool = False,
    skip_local: bool = False,
    truncate_query: Optional[int] = None,
    truncate_target: Optional[int] = None,
    force_constant: bool = False,
) -> Tuple[Optional[abutils.tl.PairwiseAlignment]]:
    """
    Performs an (optionally) two-step realignment of an assigned germline gene to a query sequence.

      * The first step is a semi-global alignment, which ensures that one or a handful of mismatches
        at the 5' or 3' end of the sequence do not result in incorrect truncation of a sequence that in fact
        should extend to the 5' or 3' end of the germline gene.

      * The second step is a local alignment, which fine-tunes the alignment with optimal alignment
        parameters and allowing truncation at the 3' (for V genes) or 5' (for J genes) end.

    Parameters
    ----------
    sequence : str
        The query sequence to align.

    germline_name : str
        The germline gene to align to.

    imgt_gapped : bool, default: False
        Whether to use the IMGT gapped or ungapped germline sequences. If ``True``, the IMGT gapped
        sequences are used.

    semiglobal_aln_params : dict, default: None
        The parameters to use for the semi-global alignment. Should be a ``dict`` containing one or
        more of the following keys:

        - ``match``: The match score (should be >= 0).
        - ``mismatch``: The mismatch score (must be <= 0).
        - ``gap_open``: The gap open penalty (must be <= 0).
        - ``gap_extend``: The gap extend penalty (must be <= 0).

    local_aln_params : dict, default: None
        The parameters to use for the local alignment. Should be a ``dict`` containing one or
        more of the following keys:

        - ``match``: The match score (should be >= 0).
        - ``mismatch``: The mismatch score (must be <= 0).
        - ``gap_open``: The gap open penalty (must be <= 0).
        - ``gap_extend``: The gap extend penalty (must be <= 0).

    skip_semiglobal : bool, default: False
        Whether to skip the semi-global alignment.

    skip_local : bool, default: False
        Whether to skip the local alignment.

    truncate_query : int, default: None
        The number of bases to truncate from the 3' end of the query sequence.

    truncate_target : int, default: None
        The number of bases to truncate from the 3' end of the target sequence.

    force_constant : bool, default: False
        Whether to force the query to match a constant germline gene. This is necessary because both IgD and diversity (D)
        gene names are formatted as IGHD, making it ambiguous whether the query is for IgD or diversity. By default,
        a supplied germline name of the format IGHD is assumed to be diversity.

    Returns
    -------
    Tuple[Optional[abutils.tl.PairwiseAlignment], Optional[abutils.tl.PairwiseAlignment]]
        The semi-global and local alignments. If ``skip_semiglobal`` is ``True``, the first element of the tuple is ``None``.
        If ``skip_local`` is ``True``, the second element of the tuple is ``None``.

    """
    # initial semi-global alignment
    germ = get_germline(
        germline_name,
        germdb_name,
        imgt_gapped=imgt_gapped,
        exact_match=True,
        force_constant=force_constant,
    )
    if truncate_query is not None:
        sequence = sequence[: -int(truncate_query)]
    if truncate_target is not None:
        germ = germ[: -int(truncate_target)]
    if skip_semiglobal:
        sg = None
    else:
        semiglobal_aln_params = (
            semiglobal_aln_params if semiglobal_aln_params is not None else {}
        )
        sg = abutils.tl.semiglobal_alignment(sequence, germ, **semiglobal_aln_params)
        sequence = sg.query[sg.query_begin : sg.query_end + 1]
        germ = sg.target[sg.target_begin : sg.target_end + 1]
    # secondary (fine-tuning) local alignment
    if skip_local:
        loc = None
    else:
        local_aln_params = local_aln_params if local_aln_params is not None else {}
        loc = abutils.tl.local_alignment(sequence, germ, **local_aln_params)
    return sg, loc


def reassign_dgene(
    sequence: str, germdb_name: str, aln_params: Optional[dict] = None
) -> abutils.tl.PairwiseAlignment:
    """
    Assigns a D gene to a query sequence using local alignment.

    Parameters
    ----------
    sequence : str
        The query sequence to align.

    germdb_name : str
        The name of the germline database.

    aln_params : dict, default: None
        The parameters to use for the local alignment.


    Returns
    -------
    abutils.tl.PairwiseAlignment
        The highest scoring local alignment. If no suitable alignment is found, ``None`` is returned.

    """
    # fetch all d-genes
    germs = get_germline("IGHD", germdb_name)
    aln_params = aln_params if aln_params is not None else {}
    # align the query sequence to all d-genes
    alns = abutils.tl.local_alignment(sequence, targets=germs, **aln_params)
    if len(alns) > 0:
        # return the highest scoring alignment
        return sorted(alns, key=lambda x: x.score, reverse=True)[0]
    return


# ------------------------------
#     ALIGNMENT PROCESSING
# ------------------------------


def process_vgene_alignment(
    semiglobal_aln: abutils.tl.PairwiseAlignment,
    local_aln: abutils.tl.PairwiseAlignment,
    ab: Antibody,
) -> Antibody:
    """
    Processes a V gene alignment and updates ``Antibody`` annotations accordingly.

    .. note:
        all start/end positions are 0-indexed and end postions are
        exclusive, which aligns with Python slicing. This means that
        ``sequence[start : end]`` will work as expected


    Parameters
    ----------
    semiglobal_aln : abutils.tl.PairwiseAlignment
        Semi-global alignment of the query sequence and a V gene.

    local_aln : abutils.tl.PairwiseAlignment
        Local alignment of the query sequence and a V gene.

    ab : Antibody
        Antibody object to update with annotation information.

        The following ``Antibody`` properties are updated:

        - ``v_sequence``: the V gene region of the query sequence
        - ``v_sequence_aa``: the V gene region of the query sequence, translated into amino acids
        - ``v_score``: the score of the local alignment
        - ``v_sequence_start``: start position of the V gene in the query sequence, in positions corresponding to the ``sequence_oriented`` property
        - ``v_sequence_end``: end position of the V gene in the query sequence, in positions corresponding to the ``sequence_oriented`` property
        - ``v_germline``: the V gene region of the assigned germline sequence
        - ``v_germline_aa``: the V gene region of the assigned germline sequence, translated into amino acids
        - ``v_germline_start``: start position of the V gene in the assigned germline sequence
        - ``v_germline_end``: end position of the V gene in the assigned germline sequence
        - ``frame``: reading frame of the aligned query sequence

        Does not require any ``Antibody`` properties to be set.

    Returns
    -------
    Antibody
        Antibody object with updated annotations.

    """
    ab.v_score = local_aln.score
    # sequence/germline start position
    if semiglobal_aln.query_begin < semiglobal_aln.target_begin:
        # the query sequence doesn't extend to the beginning of the germline
        # so we'll use the start position of the local alignment
        ab.v_sequence_start = semiglobal_aln.query_begin + local_aln.query_begin
        ab.v_germline_start = semiglobal_aln.target_begin + local_aln.target_begin
    else:
        # if the query extends 5' at least to the beginning of the germline,
        # we set the start position at the beginning of the germline
        # to avoid an edge case where a mutation at or near the start of the
        # query sequence causes the start position to be incorrect
        ab.v_sequence_start = semiglobal_aln.query_begin
        ab.v_germline_start = semiglobal_aln.target_begin
    # sequence/germline stop position
    # the +1 on v_sequence_end lets us slice nicely going forward
    # so sequence_oriented[v_sequence_start : v_sequence_end] will give the v-gene region
    ab.v_sequence_end = semiglobal_aln.query_begin + local_aln.query_end + 1
    ab.v_germline_end = semiglobal_aln.target_begin + local_aln.target_end + 1
    # v-region sequence and germline
    ab.v_sequence = semiglobal_aln.query[ab.v_sequence_start : ab.v_sequence_end]
    ab.v_germline = semiglobal_aln.target[ab.v_germline_start : ab.v_germline_end]
    # frame -- 1-indexed, which works with abutils.tl.translate()
    ab.frame = (3 - (ab.v_germline_start % 3)) % 3 + 1
    # AA sequence and germline
    ab.v_sequence_aa = abutils.tl.translate(ab.v_sequence, frame=ab.frame)
    ab.v_germline_aa = abutils.tl.translate(ab.v_germline, frame=ab.frame)
    return ab


def process_jgene_alignment(
    oriented_input: str,
    v_sequence_end: int,
    semiglobal_aln: abutils.tl.PairwiseAlignment,
    local_aln: abutils.tl.PairwiseAlignment,
    ab: Antibody,
) -> Antibody:
    """
    Processes a J gene alignment and updates ``Antibody`` annotations accordingly.

    .. note:
        all start/end positions are 0-indexed and end postions are
        exclusive, which aligns with Python slicing. This means that
        ``sequence[start : end]`` will work as expected

    Parameters:
    -----------
    oriented_input: str
        The oriented input sequence.

    v_sequence_end: int
        The end position of the V gene in the `oriented_input` sequence.

    semiglobal_aln: abutils.tl.PairwiseAlignment
        Semiglobal alignment of a query sequence to the J gene germline.

    local_aln: abutils.tl.PairwiseAlignment
        Local alignment of a query sequence to the J gene germline.

    ab: Antibody
        Antibody object to update with annotation information.

        The following ``Antibody`` properties are updated:

        - ``j_sequence``: the J gene region of the query sequence
        - ``j_score``: the score of the local alignment
        - ``j_sequence_start``: start position of the J gene in the query sequence, in positions corresponding to the ``sequence_oriented`` property
        - ``j_sequence_end``: end position of the J gene in the query sequence, in positions corresponding to the ``sequence_oriented`` property
        - ``j_germline``: the J gene region of the assigned germline sequence
        - ``j_germline_start``: start position of the J gene in the assigned germline sequence
        - ``j_germline_end``: end position of the J gene in the assigned germline sequence

        Does not require any ``Antibody`` properties to be set.

    """
    ab.j_score = local_aln.score
    # AIRR-C wants 1-based indexing, but 0-based is better so we'll do that instead
    ab.j_sequence_start = (
        v_sequence_end + semiglobal_aln.query_begin + local_aln.query_begin
    )
    ab.j_sequence_end = (
        ab.j_sequence_start + (local_aln.query_end - local_aln.query_begin) + 1
    )
    # germline start/stop positions
    ab.j_germline_start = semiglobal_aln.target_begin + local_aln.target_begin
    ab.j_germline_end = (
        ab.j_germline_start + (local_aln.target_end - local_aln.target_begin) + 1
    )
    # j-region sequence and germline
    ab.j_sequence = oriented_input[ab.j_sequence_start : ab.j_sequence_end]
    ab.j_germline = semiglobal_aln.target[ab.j_germline_start : ab.j_germline_end]
    return ab


def process_dgene_alignment(
    oriented_input: str,
    v_sequence_end: int,
    local_aln: abutils.tl.PairwiseAlignment,
    ab: Antibody,
) -> Antibody:
    """
    Processes a D gene alignment and updates ``Antibody`` annotations accordingly.

    .. note:
        all start/end positions are 0-indexed and end postions are
        exclusive, which aligns with Python slicing. This means that
        ``sequence[start : end]`` will work as expected

    Parameters:
    -----------
    oriented_input: str
        The oriented input sequence.

    v_sequence_end: int
        The end position of the V gene in the `oriented_input` sequence.

    semiglobal_aln: abutils.tl.PairwiseAlignment
        Semiglobal alignment of a query sequence to the J gene germline.

    local_aln: abutils.tl.PairwiseAlignment
        Local alignment of a query sequence to the J gene germline.

    ab: Antibody
        Antibody object to update with annotation information.

        The following ``Antibody`` properties are updated:

        - ``d_sequence``: the D gene region of the query sequence
        - ``d_score``: the score of the local alignment
        - ``d_sequence_start``: start position of the D gene in the query sequence, in positions corresponding to the ``sequence_oriented`` property
        - ``d_sequence_end``: end position of the D gene in the query sequence, in positions corresponding to the ``sequence_oriented`` property
        - ``d_germline``: the D gene region of the assigned germline sequence
        - ``d_germline_start``: start position of the D gene in the assigned germline sequence
        - ``d_germline_end``: end position of the D gene in the assigned germline sequence
        - ``d_frame``: reading frame of the D gene

        Does not require any ``Antibody`` properties to be set.

    """
    ab.d_score = local_aln.score
    # AIRR-C wants 1-based indexing, but 0-based is better so we'll do that instead
    ab.d_sequence_start = v_sequence_end + local_aln.query_begin
    d_length = local_aln.query_end - local_aln.query_begin
    ab.d_sequence_end = ab.d_sequence_start + d_length + 1
    # germline start/stop positions
    ab.d_germline_start = local_aln.target_begin
    d_germline_length = local_aln.query_end - local_aln.query_begin
    ab.d_germline_end = ab.d_germline_start + d_germline_length + 1
    # d-region sequence and germline
    ab.d_sequence = oriented_input[ab.d_sequence_start : ab.d_sequence_end]
    ab.d_germline = local_aln.target[ab.d_germline_start : ab.d_germline_end]
    # d frame
    ab.d_frame = (3 - (ab.d_germline_start % 3)) % 3 + 1
    return ab


def process_cgene_alignment(
    oriented_input: str,
    j_sequence_end: int,
    semiglobal_aln: abutils.tl.PairwiseAlignment,
    local_aln: abutils.tl.PairwiseAlignment,
    ab: Antibody,
) -> Antibody:
    """
    Processes a constant region alignment and updates ``Antibody`` annotations accordingly.

    .. note:
        all start/end positions are 0-indexed and end postions are
        exclusive, which aligns with Python slicing. This means that
        ``sequence[start : end]`` will work as expected

    Parameters:
    -----------
    oriented_input: str
        The oriented input sequence.

    j_sequence_end: int
        The end position of the J gene in the `oriented_input` sequence.

    semiglobal_aln: abutils.tl.PairwiseAlignment
        Semiglobal alignment of a query sequence to the constant region germline.

    local_aln: abutils.tl.PairwiseAlignment
        Local alignment of a query sequence to the constant region germline.

    ab: Antibody
        Antibody object to update with annotation information.

        The following ``Antibody`` properties are updated:

        - ``c_sequence``: the constant region of the query sequence
        - ``c_score``: the score of the local alignment
        - ``c_sequence_start``: start position of the constant region in the query sequence, in positions corresponding to the ``sequence_oriented`` property
        - ``c_sequence_end``: end position of the constant region in the query sequence, in positions corresponding to the ``sequence_oriented`` property
        - ``c_germline``: the constant region of the assigned germline sequence
        - ``c_germline_start``: start position of the constant region in the assigned germline sequence
        - ``c_germline_end``: end position of the constant region in the assigned germline sequence

        Does not require any ``Antibody`` properties to be set.

    """
    ab.c_score = local_aln.score
    # sequence/germline start position
    if semiglobal_aln.query_begin < semiglobal_aln.target_begin:
        # the query sequence doesn't extend to the beginning of the germline
        # so we'll use the start position of the local alignment
        ab.c_sequence_start = (
            j_sequence_end + semiglobal_aln.query_begin + local_aln.query_begin
        )
        ab.c_germline_start = semiglobal_aln.target_begin + local_aln.target_begin
    else:
        # if the query extends 5' at least to the beginning of the germline,
        # we set the start position at the beginning of the germline
        # to avoid an edge case where a mutation at or near the start of the
        # query sequence causes the start position to be incorrect
        ab.c_sequence_start = j_sequence_end + semiglobal_aln.query_begin
        ab.c_germline_start = semiglobal_aln.target_begin
    # sequence/germline stop positions
    ab.c_sequence_end = (
        ab.c_sequence_start + semiglobal_aln.query_begin + local_aln.query_end + 1
    )
    ab.c_germline_end = semiglobal_aln.target_begin + local_aln.target_end + 1
    # j-region sequence and germline
    ab.c_sequence = oriented_input[ab.c_sequence_start : ab.c_sequence_end]
    ab.c_germline = semiglobal_aln.target[ab.c_germline_start : ab.c_germline_end]
    # frame -- 1-indexed, which works with abutils.tl.translate()
    ab.c_frame = (3 - (ab.c_germline_start % 3)) % 3 + 1
    # AA sequence and germline
    ab.c_sequence_aa = abutils.tl.translate(ab.c_sequence, frame=ab.c_frame)
    ab.c_germline_aa = abutils.tl.translate(ab.c_germline, frame=ab.c_frame)
    return ab



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

import pytest

from ..annotation.antibody import Antibody
from ..annotation.mutations import annotate_mutations


@pytest.fixture
def germline():
    return "ATGCATGC"


@pytest.fixture
def aligned_sequence_with_insertion():
    return "ATGCAAAATGC"


@pytest.fixture
def aligned_germline_with_insertion():
    return "ATGC---ATGC"


@pytest.fixture
def gapped_germline():
    return "ATG.CAT...GC"


@pytest.fixture
def sequence_single_mutation_5at():
    # mutation is A>T at raw position 5
    return "ATGCTTGC"


@pytest.fixture
def sequence_single_mutation_7gc():
    # mutation is G>C at raw position 7
    return "ATGCATCC"


@pytest.fixture
def sequence_multi_mutation_5at_7gc():
    # mutations (raw positions) are 5:A>T and 7:G>C
    return "ATGCTTCC"


def test_annotate_mutations_no_mutations(germline):
    mutations = annotate_mutations(
        aligned_sequence=germline,
        aligned_germline=germline,
        gapped_germline=germline,
        germline_start=0,
    )
    assert len(mutations) == 0


def test_annotate_mutations_single_mutation_5at(germline, sequence_single_mutation_5at):
    mutations = annotate_mutations(
        aligned_sequence=sequence_single_mutation_5at,
        aligned_germline=germline,
        gapped_germline=germline,
        germline_start=0,
    )
    assert "5:A>T" in mutations
    assert len(mutations) == 1


def test_annotate_mutations_single_mutation_7gc(germline, sequence_single_mutation_7gc):
    mutations = annotate_mutations(
        aligned_sequence=sequence_single_mutation_7gc,
        aligned_germline=germline,
        gapped_germline=germline,
        germline_start=0,
    )
    assert "7:G>C" in mutations
    assert len(mutations) == 1


def test_annotate_mutations_single_mutation_single_germline_gap(
    germline, sequence_single_mutation_5at, gapped_germline
):
    mutations = annotate_mutations(
        aligned_sequence=sequence_single_mutation_5at,
        aligned_germline=germline,
        gapped_germline=gapped_germline,
        germline_start=0,
    )
    assert "6:A>T" in mutations
    assert len(mutations) == 1


def test_annotate_mutations_single_mutation_multi_germline_gap(
    germline, sequence_single_mutation_7gc, gapped_germline
):
    mutations = annotate_mutations(
        aligned_sequence=sequence_single_mutation_7gc,
        aligned_germline=germline,
        gapped_germline=gapped_germline,
        germline_start=0,
    )
    assert "11:G>C" in mutations
    assert len(mutations) == 1


def test_annotate_mutations_multi_mutation(germline, sequence_multi_mutation_5at_7gc):
    mutations = annotate_mutations(
        aligned_sequence=sequence_multi_mutation_5at_7gc,
        aligned_germline=germline,
        gapped_germline=germline,
        germline_start=0,
    )
    assert "5:A>T" in mutations
    assert "7:G>C" in mutations
    assert len(mutations) == 2


def test_annotate_mutations_multi_mutation_germline_gaps(
    germline, sequence_multi_mutation_5at_7gc, gapped_germline
):
    mutations = annotate_mutations(
        aligned_sequence=sequence_multi_mutation_5at_7gc,
        aligned_germline=germline,
        gapped_germline=gapped_germline,
        germline_start=0,
    )
    assert "6:A>T" in mutations
    assert "11:G>C" in mutations
    assert len(mutations) == 2


def test_annotate_mutations_single_mutation_5at_aa(
    germline, sequence_single_mutation_5at
):
    mutations = annotate_mutations(
        aligned_sequence=sequence_single_mutation_5at,
        aligned_germline=germline,
        gapped_germline=germline,
        germline_start=0,
    )
    assert "5:A>T" in mutations
    assert len(mutations) == 1



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

import pytest
from abutils import Sequence

from ..annotation.germline import get_germline, get_germline_database_path

# ----------------------------
#      DATABASE PATHS
# ----------------------------


def test_get_germline_database_path():
    path = get_germline_database_path(germdb_name="human", receptor="bcr")
    assert path


@pytest.mark.xfail(
    reason="a database named NotAGermlineDatabase does not exist",
    raises=FileNotFoundError,
)
def test_get_germline_database_name_invalid():
    path = get_germline_database_path(
        germdb_name="NotAGermlineDatabase", receptor="bcr"
    )
    assert path


@pytest.mark.xfail(
    reason="receptor type 'abc' is invalid (must be 'bcr' or 'tcr')",
    raises=ValueError,
)
def test_get_germline_database_receptor_invalid():
    path = get_germline_database_path(germdb_name="human", receptor="abc")
    assert path


# ----------------------------
#        GET GERMLINE
# ----------------------------


def test_get_single_germline():
    germ = get_germline(
        germline_gene="IGHV1-2*02",
        germdb_name="human",
        receptor="bcr",
        exact_match=True,
    )
    assert isinstance(germ, Sequence)
    assert germ.id == "IGHV1-2*02"


def test_get_multiple_germlines():
    germs = get_germline(
        germline_gene="IGHV1-2",
        germdb_name="human",
        receptor="bcr",
        exact_match=False,
    )
    assert len(germs) >= 2
    assert all([isinstance(germ, Sequence) for germ in germs])
    assert all([germ.id.startswith("IGHV1-2") for germ in germs])


def test_get_single_germline_tcr():
    germ = get_germline(
        germline_gene="TRAV1-1*01",
        germdb_name="human",
        receptor="tcr",
        exact_match=True,
    )
    assert isinstance(germ, Sequence)
    assert germ.id == "TRAV1-1*01"


def test_get_multiple_germlines_tcr():
    germs = get_germline(
        germline_gene="TRAV1-1",
        germdb_name="human",
        receptor="tcr",
        exact_match=False,
    )
    assert len(germs) >= 2
    assert all([isinstance(germ, Sequence) for germ in germs])
    assert all([germ.id.startswith("TRAV1-1") for germ in germs])


@pytest.mark.xfail(
    reason="gene 'IGHV1-1*01' does not exist in the human bcr database",
    raises=ValueError,
)
def test_get_single_germline_nonexistent():
    germ = get_germline(
        germline_gene="IGHV1-1*01",
        germdb_name="human",
        receptor="bcr",
        exact_match=True,
    )
    assert isinstance(germ, Sequence)
    assert germ.id == "IGHV1-1*01"


@pytest.mark.xfail(
    reason="gene 'IGHV1-2' does not identically match anything in the human bcr database",
    raises=ValueError,
)
def test_get_single_germline_nonunique():
    germ = get_germline(
        germline_gene="IGHV1-2",
        germdb_name="human",
        receptor="bcr",
        exact_match=True,
    )
    assert isinstance(germ, Sequence)
    assert germ.id == "IGHV1-2"



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

import os
import shutil
import tempfile
from typing import Iterable, Optional, Union

import abutils
from abutils import Sequence
from abutils.tl import PairwiseAlignment

__all__ = ["parse_umis"]


class UMI:
    """
    Class for parsing UMIs from sequence data.
    """

    def __init__(
        self,
        sequence: Sequence,
        pattern: str,
        length: Optional[int] = None,
        ignore_strand: bool = False,
        # extra_length_for_alignment: int = 25,
    ):
        """
        Parameters
        ----------
        sequence : str or Sequence
            Raw input sequence

        pattern : str
            Pattern for identifying the UMI

        length : int, default=None
            Length of the UMI

        ignore_strand : bool, default=False
            Ignore the strand of the sequence

        extra_length_for_alignment : int, default=25
            Extra length added to the query sequence during pattern alignment. This allows for some ambiguity
            in the location of the UMI without needing to align the complete sequence (saving compute time).

        """
        self.length = length
        self.pattern = pattern if pattern is None else pattern.strip().upper()
        self.ignore_strand = ignore_strand
        # self.extra_length_for_alignment = extra_length_for_alignment
        self.sequence = self.process_sequence(sequence)
        self._leading_aln = None
        self._trailing_aln = None
        self._umi = None
        self._num_mismatches = None

    @property
    def leading(self) -> Optional[str]:
        """
        Returns the leading conserved region of the pattern, if present.
        """
        if self.pattern is not None:
            leading = self.pattern.split("[UMI]")[0]
            return leading if leading else None

    @property
    def trailing(self) -> Optional[str]:
        """
        Returns the trailing conserved region of the pattern, if present.
        """
        if self.pattern is not None:
            trailing = self.pattern.split("[UMI]")[-1]
            return trailing if trailing else None

    @property
    def leading_aln(self) -> Optional[PairwiseAlignment]:
        """
        Returns the leading conserved region of the pattern, if present.
        """
        if self.pattern is not None:
            if self._leading_aln is None:
                if self.leading is not None:
                    self._leading_aln = self.align(self.leading, self.sequence)
            return self._leading_aln

    @property
    def trailing_aln(self) -> Optional[PairwiseAlignment]:
        """
        Returns the trailing conserved region of the pattern, if present.
        """
        if self.pattern is not None:
            if self._trailing_aln is None:
                if self.trailing is not None:
                    self._trailing_aln = self.align(self.trailing, self.sequence)
            return self._trailing_aln

    @property
    def umi(self) -> Optional[str]:
        """
        Returns the UMI, if present.
        """
        if self._umi is None:
            self._umi = self.get_umi()
        return self._umi

    @property
    def num_mismatches(self) -> Optional[int]:
        """
        Returns the number of mismatches between the pattern and sequence.
        """
        if self._num_mismatches is None:
            self._num_mismatches = self.get_mismatches()
        return self._num_mismatches

    def process_sequence(self, sequence: Sequence) -> str:
        """
        Processes the raw sequence:
          * compute the reverse complement if length is negative
          * truncate sequence prior to pattern alignment

        Parameters
        ----------
        sequence : str or Sequence
            Raw sequence

        Returns
        -------
        truncated : str
        """
        if self.pattern is not None:
            # length_for_aln = (
            #     len(self.pattern.replace("[UMI]", "")) + self.extra_length_for_alignment
            # )
            # length_for_aln += self.length if self.length is not None else 16
            # sequence = sequence[:length_for_aln]
            if self.length < 0 and not self.ignore_strand:
                sequence = abutils.tl.reverse_complement(sequence)
        return sequence

    def align(self, pattern: str, sequence: Sequence) -> PairwiseAlignment:
        """
        Performs local alignment between a pattern and sequence.
        """
        aln = abutils.tl.local_alignment(
            pattern, sequence, gap_open=-20, gap_extend=-20
        )
        return aln

    def get_mismatches(self) -> Optional[int]:
        """
        Returns the total number of mismatches between the pattern(s)
        and the input sequence.
        """
        if self.pattern is None:
            return 0
        total = len(self.pattern.replace("[UMI]", ""))
        matches = 0
        for aln in [self.leading_aln, self.trailing_aln]:
            if aln is not None:
                matches += aln.alignment_midline.count("|")
        return total - matches

    def get_umi(self) -> Optional[str]:
        """
        Parses UMI from the input sequence using the UMI pattern(s).
        """
        # length without pattern means we just slice
        if self.pattern is None:
            if self.length < 0:
                return self.sequence[self.length :]
            else:
                return self.sequence[: self.length]
        ## something like "ATGC[UMI]"
        ## we need to take `length` residues following the end of the leading alignment
        if self.trailing is None and self.leading is not None:
            start = self.leading_aln.target_end + 1
            end = start + self.length
            if end > len(self.sequence):
                return None
        ## something like "[UMI]ATGC"
        ## we need to take `length` residues preceeding the start of the trailing alignment
        elif self.leading is None and self.trailing is not None:
            end = self.trailing_aln.target_begin
            if self.length is not None:
                start = end - self.length
            else:
                start = 0
            ## if self.length is longer than the region preceeding the start of the
            ## trailing alignment, return None because we can't parse a complete UMI
            if start < 0:
                return None
        else:
            start = self.leading_aln.target_end + 1
            end = self.trailing_aln.target_begin
        return self.sequence[start:end]


def parse_umis(
    sequences: Union[str, Sequence, Iterable, dict],
    output_file: Optional[str] = None,
    pattern: Union[str, Iterable, None] = None,
    length: Union[int, Iterable, None] = None,
    allowed_mismatches: int = 1,
    # extra_length_for_alignment: int = 25,
    ignore_strand: bool = False,
    fmt: str = "fasta",
    sequence_key: str = "sequence_input",
    id_key: str = "sequence_id",
) -> str:
    """
    Parses unique molecular identifiers (UMIs) from bulk NGS sequencing data.

    Parameters
    ----------
    sequences : str or iterable
        Can be one of several things:
            1. path to a FASTA- or FASTQ-formatted file
            2. a single sequence, as a string or a ``Sequence`` object
            3. a list of ``Sequence`` objects or anything accepted by ``abutils.Sequence()``

    output_file : str, default=None
        Path to an output file. If `sequences` is not a file, and `output_file` is not
        provided `sequences` are converted to ``abutils.Sequence`` objects and the UMI
        is added to the ``"umi"`` field of ``sequence.annotations``. If `sequences` is
        a file and `output` is not provided, UMI parsing is done in-place and the
        `sequences` file is updated. UMIs are added to the end of each sequence ID with
        an underscore (``"_"``) separating the name and UMI. Sequences for which at least
        one UMI was not found are not included in the output.

    pattern : str or iterable, default=None
        Pattern (or iterable of patterns) for identifying the location of the UMI,
        or the name of a built-in pattern. Built-in options include ``"smartseq-human-bcr"``.
        Patterns may optionally contain leading and/or trailing conserved regions, with
        the UMI position within the pattern represented by ``"[UMI]"``. As an example,
        one of the built-in patterns for SmartSeq-BCR UMIs is::

            "[UMI]TCAGCGGGAAGACATT"

        which would be a UMI sequence followed immediately by ``"TCAGCGGGAAGACATT"``.
        By default, the pattern is matched to the 5' -> 3' strand. This allows
        users to more easily construct patterns from their amplification primers without
        needing to worry about reverse-complementing patterns for UMIs at the 3' end of the
        input sequence. To override this, set `ignore_strand` to ``True``. If `pattern`
        is not provided, UMIs will be parsed using only `length`, starting at the start
        (if `length` is positive) or end (if `length` is negative) of the sequence.

        .. note::

            The UMIs for all `patterns` that meet the `allowed_mismatches` criteria in
            the conserved leading/trailing portions will be concatenated into the final
            UMI. This allows the use of multiple `patterns` for either:
              - different patterns designed to match one of several different types of
                sequences in a heterogeneous sample. For example, if heavy, kappa and
                lambda primers each have different conserved regions flanking the UMI
                and the input file contains a mix of heavy, kappa and lambda chains,
                supplying all patterns (assuming they're sufficiently different from
                each other) will allow parsing of UMIs from all chains.
              - different patterns for sequences that contain multiple UMIs, either
                at opposite ends of the sequence or on the same end of the sequence
                but in different locations and with different conserved flanking regions.

        If multiple UMIs are parsed from a single sequence, they will be concatenated with
        ``"+"`` separating the UMIs.

    length : int or iterable, default=None
        Length of the UMI sequence, or iterable of UMI lengths. If `length` is provided
        and pattern is not, the UMI will be sliced directly from the start (if `length` is
        positive) or the end (if `length` is negative) of the sequencing without attempting
        to match UMI-flanking sequence patterns. If multiple lengths are
        provided, there must be an equal number of `patterns`, and they should be in the
        same order (the first `pattern` should correspond to the first UMI `length`). If
        `length` is positive, the UMI will be parsed from the start of the sequence. If
        `length` is negative, the UMI will be parsed from the end of the sequence. If multiple
        `patterns` are provided with a single `length`, that `length` will be used for all
        `patterns`. Required if `pattern` does not have a conserved trailing region.
        If `length` is not provided and `pattern` has a trailing conserved region but not a
        leading conserved region, the entire portion of `sequence` preceeding the trailing
        region will be parsed as the UMI. Ignored if both leading and trailing sequences
        are present in `pattern`, as the entire region between the conserved flanking regions
        will be parsed as the UMI regardless of what `length` is provided.

    allowed_mismatches : int, default=None
        Total number of mismatches allowed when aligning the conserved flanking regions
        of `pattern`. Mismatches are calculated by subtracting the number of identically
        matching alignment positions from the total length of the conserved flanking region.
        If not provided, the default is to allow ``1`` mismatch.

    extra_length_for_alignment : int, default=25
        To speed alignment and avoid coincidental matches in regions of the sequence that are
        not logical locations for UMIs, only a fragment of each input sequences (sliced from
        the start or end of the seqeunce, depending on `length`) is aligned against conserved
        `pattern` regions. The size of this fragment is computed by summing the total length of
        conserved flanking regions, the UMI `length`, and `extra_length_for_alignment`. The default
        is ``25``, which should be sufficient for most cases, however, `extra_length_for_alignment`
        may need to be increased if the UMI is not located close to the start or end of the
        input seqeunce.

    ignore_strand : bool, default=False
        If ``True``, patterns matching to the end of the input sequence (those with a negative
        `length`) will be aligned to the unmodified input sequence rather than the reverse
        complement. This means that the supplied `pattern` must be in the 3' -> 5' orientation.

    fmt : str, default="fasta"
        Format of the output file. Only used if `output_file` is ``True``. Options are ``"fasta"``
        and ``"fastq"``.

        .. warning::

            If selecting ``"fastq"`` output when providing a FASTA-formatted input file,
            the resulting FASTQ file may be malformed as there is no quality information
            in the input file.

    sequence_key : str, default="raw_input"
        Field in `sequences` from which the UMI should be parsed. Only used if `sequences`
        are ``abutils.Sequence`` objects. Default is ``"raw_input"``.  If `sequence_key` is
        not found, or is ``None`` ``sequence.sequence`` will be used.

    id_key : str, default="sequence_id"
        Field in `sequences` containing the sequence ID. Only used if `sequences` are
        ``abutils.Sequence`` objects. Default is ``"sequence_id"``. If `id_key` is not found
        or is ``None`` ``sequence.id`` will be used.


    Returns
    -------
    output_file : str
        Path to the output file, with a format (``"fasta"`` or ``"fastq"``) matching `fmt`.

    """
    # lengths and patterns
    if pattern is None:
        if isinstance(length, int):
            patterns = [None]
            lengths = [length]
        else:
            patterns = [None] * len(length)
            lengths = length
    elif pattern is not None:
        if isinstance(pattern, str):
            if pattern.lower() in BUILTIN_PATTERNS:
                name = pattern.lower()
                patterns = _get_patterns(name)
                length = _get_lengths(name)
                if allowed_mismatches is None:
                    allowed_mismatches = _get_allowed_mismatches(name)
            else:
                patterns = [pattern]
        else:
            patterns = pattern
        if isinstance(length, int) or length is None:
            lengths = [length] * len(patterns)
        else:
            lengths = length
        if len(patterns) != len(lengths):
            err = "\nERROR: if pattern and length are both iterables,"
            err += " they must be the same length.\n"
            err += f"  - pattern has {len(patterns)} elements: {', '.join(patterns)}\n"
            err += f"  - length has {len(lengths)} elements: {', '.join(lengths)}\n"
            raise ValueError(err)
    # validate format
    fmt = fmt.lower()
    if fmt not in ["fasta", "fastq"]:
        err = "\nERROR: format must be either 'fasta' or 'fastq'. "
        err += f"You provided {fmt}.\n"
        raise ValueError(err)
    # input processing
    if isinstance(sequences, str) and os.path.isfile(sequences):
        output = _parse_umis_from_file(
            input_file=sequences,
            patterns=patterns,
            lengths=lengths,
            output_file=output_file,
            allowed_mismatches=allowed_mismatches,
            # extra_length_for_alignment=extra_length_for_alignment,
            ignore_strand=ignore_strand,
            fmt=fmt,
        )
    elif isinstance(sequences, (str, Sequence)):
        output = _parse_umis_from_single_sequence(
            sequence=sequences,
            patterns=patterns,
            lengths=lengths,
            allowed_mismatches=allowed_mismatches,
            # extra_length_for_alignment=extra_length_for_alignment,
            ignore_strand=ignore_strand,
        )
    else:
        output = _parse_umis_from_sequences(
            sequences=sequences,
            patterns=patterns,
            lengths=lengths,
            output_file=output_file,
            allowed_mismatches=allowed_mismatches,
            # extra_length_for_alignment=extra_length_for_alignment,
            ignore_strand=ignore_strand,
            fmt=fmt,
            sequence_key=sequence_key,
            id_key=id_key,
        )
    return output


def _parse_umis_from_file(
    input_file: str,
    patterns: Iterable,
    lengths: Iterable,
    output_file: Optional[str] = None,
    allowed_mismatches: int = 1,
    # extra_length_for_alignment: int = 25,
    ignore_strand: bool = False,
    separator: str = "_",
    fmt: str = "fasta",
) -> str:
    """
    Parses UMIs from a FASTA- or FASTQ-formatted file.
    """
    # set up output file
    inplace = False
    if output_file is None:
        output_file = tempfile.NamedTemporaryFile(delete=False).name
        inplace = True
    else:
        abutils.io.make_dir(os.path.dirname(output_file))
    with open(output_file, "w") as ofile:
        for s in abutils.io.parse_fastx(input_file):
            # with open(input_file, "r") as ifile:
            #     for s in SeqIO.parse(ifile, fmt.lower()):
            #         s = abutils.Sequence(s)
            umi_list = []
            for pattern, length in zip(patterns, lengths):
                u = UMI(
                    sequence=s,
                    pattern=pattern,
                    length=length,
                    # extra_length_for_alignment=extra_length_for_alignment,
                    ignore_strand=ignore_strand,
                )
                if u.num_mismatches <= allowed_mismatches:
                    umi_list.append(u.umi)
            # if we couldn't find a UMI, skip the sequence
            umi_list = [u for u in umi_list if u is not None]
            if not umi_list:
                continue
            umi = "+".join(umi_list)
            s.id = f"{s.id}{separator}{umi}"
            if fmt.lower() == "fastq":
                fstring = s.fastq + "\n"
            else:
                fstring = s.fasta + "\n"
            ofile.write(fstring)
    if inplace:
        output_file = shutil.move(output_file, input_file)
    return output_file


def _parse_umis_from_sequences(
    sequences: Iterable,
    patterns: Iterable,
    lengths: Iterable,
    output_file: str,
    allowed_mismatches: int = 1,
    # extra_length_for_alignment: int = 25,
    ignore_strand: bool = False,
    fmt: str = "fasta",
    sequence_key: str = "sequence_input",
    id_key: str = "sequence_id",
) -> str:
    """
    Parses UMIs from a list of sequences.
    """
    inplace = True if output_file is None else False
    # output is a list of either abutils.Sequence objects (if inplace == True)
    # or a list of FAST[AQ]-formatted strings (if inplace == False)
    output = []
    for s in sequences:
        if isinstance(s, Sequence):
            if sequence_key in s:
                s.sequence = s[sequence_key]
            if id_key in s:
                s.id = s[id_key]
        else:
            s = Sequence(s)
        umi_list = []
        for pattern, length in zip(patterns, lengths):
            u = UMI(
                sequence=s,
                pattern=pattern,
                length=length,
                # extra_length_for_alignment=extra_length_for_alignment,
                ignore_strand=ignore_strand,
            )
            if u.num_mismatches <= allowed_mismatches:
                umi_list.append(u.umi)
        # if we couldn't find a UMI, skip the sequence
        umi_list = [u for u in umi_list if u is not None]
        if not umi_list:
            continue
        umi = "+".join(umi_list)
        if inplace:
            s["umi"] = umi
            output.append(s)
        else:
            s.id = f"{s.id}_{umi}"
            if fmt.lower() == "fastq":
                output.append(s.fastq)
            else:
                output.append(s.fasta)
    if inplace:
        return output
    else:
        with open(output_file, "w") as ofile:
            ofile.write("\n".join(output))
        return output_file


def _parse_umis_from_single_sequence(
    sequence: Union[str, Sequence, Iterable, dict],
    patterns: Iterable,
    lengths: Iterable,
    allowed_mismatches: int = 1,
    # extra_length_for_alignment: int = 25,
    ignore_strand: bool = False,
    sequence_key: str = "sequence_input",
    id_key: str = "sequence_id",
) -> str:
    """
    Parses UMIs from a single sequence.
    """
    # process input
    s = Sequence(sequence)
    if sequence_key in s:
        s.sequence = s[sequence_key]
    if id_key in s:
        s.id = s[id_key]
    # find UMIs
    umi_list = []
    for pattern, length in zip(patterns, lengths):
        u = UMI(
            sequence=s,
            pattern=pattern,
            length=length,
            # extra_length_for_alignment=extra_length_for_alignment,
            ignore_strand=ignore_strand,
        )
        if u.num_mismatches <= allowed_mismatches:
            umi_list.append(u.umi)
    # concatenate all of the UMI segments with "+"
    umi = "+".join([u for u in umi_list if u is not None])
    # if we couldn't find a UMI (UMI is an empty string), return None
    if not umi:
        return None
    return umi


def _get_patterns(name):
    p = BUILTIN_PATTERNS.get(name, None)
    if p is not None:
        return p.get("pattern", None)


def _get_lengths(name):
    p = BUILTIN_PATTERNS.get(name, None)
    if p is not None:
        return p.get("length", None)


def _get_allowed_mismatches(name):
    p = BUILTIN_PATTERNS.get(name, None)
    if p is not None:
        return p.get("allowed_mismatches", None)


# ---------------------------
#        PATTERNS
# ---------------------------

# Takara's Smart-Seq Human BCR kit
# https://www.takarabio.com/products/next-generation-sequencing/immune-profiling/human-repertoire/smart-seq-human-bcr-with-umis
SMARTSEQ_HUMAN_BCR_PATTERN = [
    "[UMI]TCAGCGGGAAGACATT",
    "[UMI]GGGCGGATGGACTACC",
    "[UMI]CCGATGGGCGCTTGGT",
    "[UMI]GGAACACATGCGGAGC",
]
SMARTSEQ_HUMAN_BCR_LENGTH = 12
SMARTSEQ_HUMAN_BCR_MISMATCH = 2


BUILTIN_PATTERNS = {
    "smartseq-human-bcr": {
        "pattern": SMARTSEQ_HUMAN_BCR_PATTERN,
        "length": SMARTSEQ_HUMAN_BCR_LENGTH,
        "allowed_mismatches": SMARTSEQ_HUMAN_BCR_MISMATCH,
    },
}


# import os
# import shutil
# import sys
# import tempfile
# from typing import Union, Iterable, Optional

# from Bio import SeqIO
# from Bio.SeqRecord import SeqRecord

# import abutils


# class UMI:
#     """ """

#     def __init__(
#         self,
#         sequence: Union[str, abutils.Sequence, SeqRecord],
#         pattern: str,
#         length: Optional[int] = None,
#         ignore_strand: bool = False,
#         extra_length_for_alignment: int = 25,
#     ):
#         """ """
#         self.length = length
#         self.pattern = pattern.strip().upper()
#         self.ignore_strand = ignore_strand
#         self.extra_length_for_alignment = extra_length_for_alignment
#         self.sequence = self.process_sequence(sequence)
#         self._leading_aln = None
#         self._trailing_aln = None
#         self._umi = None
#         self._num_mismatches = None

#     @property
#     def leading(self) -> Optional[str]:
#         leading = self.pattern.split("[UMI]")[0]
#         return leading if leading else None

#     @property
#     def trailing(self) -> Optional[str]:
#         trailing = self.pattern.split("[UMI]")[-1]
#         return trailing if trailing else None

#     @property
#     def leading_aln(self) -> abutils.SSWAlignment:
#         if self._leading_aln is None:
#             if self.leading is not None:
#                 self._leading_aln = self.align(self.leading, self.sequence)
#         return self._leading_aln

#     @property
#     def trailing_aln(self) -> abutils.SSWAlignment:
#         if self._trailing_aln is None:
#             if self.trailing is not None:
#                 self._trailing_aln = self.align(self.trailing, self.sequence)
#         return self._trailing_aln

#     @property
#     def umi(self) -> Optional[str]:
#         if self._umi is None:
#             self._umi = self.get_umi()
#         return self._umi

#     @property
#     def num_mismatches(self) -> int:
#         """
#         Returns the number of mismatches between the pattern and sequence.
#         """
#         if self._num_mismatches is None:
#             self._num_mismatches = self.get_mismatches()
#         return self._num_mismatches

#     def process_sequence(self, sequence) -> str:
#         """
#         Processes the raw sequence:
#           * reverse complement if length is negative
#           * truncate sequence prior to pattern alignment

#         Parameters
#         ----------
#         sequence : str, Sequence, or SeqRecord
#             Raw sequence

#         Returns
#         -------
#         truncated : str
#         """
#         s = abutils.Sequence(sequence)
#         length_for_aln = (
#             len(self.pattern.replace("[UMI]", "")) + self.extra_length_for_alignment
#         )
#         if self.length is not None:
#             length_for_aln += self.length
#         if self.length < 0 and not self.ignore_strand:
#             return s.reverse_complement[:length_for_aln]
#         else:
#             return s[:length_for_aln]

#     def align(
#         self, pattern: str, sequence: Union[abutils.Sequence, str]
#     ) -> abutils.SSWAlignment:
#         """
#         Performs local alignment between a pattern and sequence.
#         """
#         aln = abutils.tl.local_alignment(
#             pattern, sequence, gap_open=-20, gap_extend=-20
#         )
#         return aln

#     def get_mismatches(self) -> Optional[int]:
#         """
#         Returns the total number of mismatches between the pattern(s)
#         and the input sequence.
#         """
#         total = len(self.pattern.replace("[UMI]", ""))
#         matches = 0
#         for aln in [self.leading_aln, self.trailing_aln]:
#             if aln is not None:
#                 matches += aln.alignment_midline.count("|")
#         return total - matches

#     def get_umi(self) -> Optional[str]:
#         """
#         Parses UMI from the input sequence using the UMI pattern(s).
#         """
#         ## something like "ATGC[UMI]"
#         ## we need to take `length` residues following the end of the leading alignment
#         if self.trailing is None and self.leading is not None:
#             start = self.leading_aln.target_end + 1
#             end = start + self.length
#             if end > len(self.sequence):
#                 return None
#         ## something like "[UMI]ATGC"
#         ## we need to take `length` residues preceeding the start of the trailing alignment
#         elif self.leading is None and self.trailing is not None:
#             end = self.trailing_aln.target_begin
#             if self.length is not None:
#                 start = end - self.length
#             else:
#                 start = 0
#             ## if self.length is longer than the region preceeding the start of the
#             ## trailing alignment, return None because we can't parse a complete UMI
#             if start < 0:
#                 return None
#         else:
#             start = self.leading_aln.target_end + 1
#             end = self.trailing_aln.target_begin
#         return self.sequence[start:end]


# def parse_umis(
#     sequences: Union[str, Iterable],
#     output_file: Optional[str] = None,
#     pattern: Union[str, Iterable, None] = None,
#     length: Union[int, Iterable, None] = None,
#     allowed_mismatches: Optional[int] = None,
#     extra_length_for_alignment: int = 25,
#     ignore_strand: bool = False,
#     fmt: str = "fasta",
#     sequence_key: str = "raw_input",
#     id_key: str = "sequence_id",
# ):
#     """
#     Parses unique molecular identifiers (UMIs) from bulk NGS sequencing data.

#     Parameters
#     ----------
#     sequences : str or iterable
#         Can be one of several things:
#             1. path to a FASTA- or FASTQ-formatted file
#             2. a list of BioPython ``SeqRecord`` objects
#             3. a list of abutils ``Sequence`` objects
#             4. a list of lists/tuples, of the format ``[sequence_id, sequence]``

#     output_file : str, default=None
#         Path to an output file. Required if `sequences` is not a
#         file. If `sequences` is a file and `output` is not provided, UMI parsing is
#         done in-place and the `sequences` file is updated.

#     pattern : str or iterable, default=None
#         Pattern (or iterable of patterns) for identifying the location of the UMI,
#         or the name of a built-in pattern. Built-in options include ``"smartseq-human-bcr"``.
#         Patterns may optionally contain leading and/or trailing conserved regions, with
#         the UMI position within the pattern represented by ``"[UMI]"``. As an example,
#         the built-in pattern for SmartSeq-BCR UMIs is::

#             "[UMI]TCAGCGGGAAGACATT"

#         which would be a UMI sequence followed immediately by ``"TCAGCGGGAAGACATT"``.
#         By default, the pattern is matched on the 5' -> 3' strand. This allows
#         users to more easily construct patterns from their amplification primers without
#         needing to worry about reverse-complementing patterns for UMIs at the 3' end of the
#         input sequence. To override this, set `ignore_strand` to ``True``. If `pattern`
#         is not provided, UMIs will be parsed using only `length`, starting at the start
#         or end of the sequence.

#         .. note::

#             The UMIs for all `patterns` that meet the `allowed_mismatches` criteria in
#             the conserved leading/trailing portions will be concatenated into the final
#             UMI. This allows the use of multiple `patterns` for either:
#               * different patterns designed to match one of several different types of
#                 sequences in a heterogeneous sample. For example, if heavy, kappa and
#                 lambda primers each have different conserved regions flanking the UMI
#                 and the input file contains a mix of heavy, kappa and lambda chains,
#                 supplying all patterns (assuming they're sufficiently different from
#                 each other) will allow parsing of UMIs from all chains.
#               * different patterns for sequences that contain multiple UMIs, either
#                 at opposite ends of the sequence or on the same end of the sequence
#                 but in different locations and with different conserved flanking regions.

#     length : int or iterable, default=None
#         Length of the UMI sequence, or iterable of UMI lengths. If multiple lengths are
#         provided, there must be an equal number of `patterns`, and they must be in the
#         same order (the first `pattern` should correspond to the first UMI `length`). If
#         `length` is positive, the UMI will be parsed from the start of the sequence. If
#         `length` is negative, the UMI will be parsed from the end of the sequence. If multiple
#         `patterns` are provided with a single `length`, that `length` will be used for all
#         `patterns`. Required if `pattern` does not have a conserved trailing region.
#         If `length` is not provided and a trailing conserved region is present in `pattern`,
#         the entire portion of `sequence` preceeding the trailing region will be parsed as the
#         UMI. Ignored if both leading and trailing sequences are present in `pattern`,
#         as the entire region between the conserved flanking regions will be parsed as
#         the UMI regardless of what `length` is provided.


#     """
#     # lengths and patterns
#     if isinstance(pattern, str):
#         if pattern.lower() in BUILTIN_PATTERNS:
#             name = pattern.lower()
#             patterns = _get_patterns(name)
#             length = _get_lengths(name)
#             if allowed_mismatches is None:
#                 allowed_mismatches = _get_allowed_mismatches(name)
#         else:
#             patterns = [pattern]
#     else:
#         patterns = pattern
#     if isinstance(length, int) or length is None:
#         lengths = [length] * len(patterns)
#     else:
#         lengths = length
#     if len(patterns) != len(lengths):
#         err = "\nERROR: if pattern and length are both iterables,"
#         err += " they must be the same length.\n"
#         err += f"  - pattern has {len(patterns)} elements: {', '.join(patterns)}\n"
#         err += f"  - length has {len(lengths)} elements: {', '.join(lengths)}\n"
#         print(err)
#         sys.exit()
#     # mismatches
#     if allowed_mismatches is None:
#         allowed_mismatches = 1
#     # validate format
#     fmt = fmt.lower()
#     if fmt not in ["fasta", "fastq"]:
#         err = "\nERROR: format must be either 'fasta' or 'fastq'. "
#         err += f"You provided {fmt}.\n"
#         print(err)
#         sys.exit()
#     # input processing
#     if isinstance(sequences, str) and os.path.isfile(sequences):
#         output_file = _parse_umis_from_file(
#             input_file=sequences,
#             patterns=patterns,
#             lengths=lengths,
#             output_file=output_file,
#             allowed_mismatches=allowed_mismatches,
#             extra_length_for_alignment=extra_length_for_alignment,
#             ignore_strand=ignore_strand,
#             fmt=fmt,
#         )
#     else:
#         if output_file is None:
#             err = "\nERROR: if input is not a FASTA- or FASTQ-formatted file, "
#             err += f"output_file must be provided.\n"
#             print(err)
#             sys.exit()
#         output_file = _parse_umis_from_sequences(
#             sequences=sequences,
#             patterns=patterns,
#             lengths=lengths,
#             output_file=output_file,
#             allowed_mismatches=allowed_mismatches,
#             extra_length_for_alignment=extra_length_for_alignment,
#             ignore_strand=ignore_strand,
#             fmt=fmt,
#             sequence_key=sequence_key,
#             id_key=id_key,
#         )


# def _parse_umis_from_file(
#     input_file: str,
#     patterns: Iterable,
#     lengths: Iterable,
#     output_file: Optional[str] = None,
#     allowed_mismatches: int = 1,
#     extra_length_for_alignment: int = 25,
#     ignore_strand: bool = False,
#     fmt: str = "fasta",
# ) -> str:
#     """
#     Parses UMIs from a FASTA- or FASTQ-formatted file.
#     """
#     # set up output file
#     inplace = False
#     if output_file is None:
#         output_file = tempfile.NamedTemporaryFile(delete=False).name
#         inplace = True
#     with open(output_file, "w") as ofile:
#         with open(input_file, "r") as ifile:
#             for s in SeqIO.parse(ifile, fmt.lower()):
#                 s = abutils.Sequence(s)
#                 umi_list = []
#                 for p, l in zip(patterns, lengths):
#                     u = UMI(
#                         sequence=s,
#                         pattern=p,
#                         length=l,
#                         extra_length_for_alignment=extra_length_for_alignment,
#                         ignore_strand=ignore_strand,
#                     )
#                     if u.num_mismatches <= allowed_mismatches:
#                         umi_list.append(u.umi)
#                 # if we couldn't find a UMI, skip the sequence
#                 umi_list = [u for u in umi_list if u is not None]
#                 if not umi_list:
#                     continue
#                 umi = "+".join(umi_list)
#                 s.id = f"{s.id}_{umi}"
#                 if fmt.lower() == "fastq":
#                     fstring = s.fastq + "\n"
#                 else:
#                     fstring = s.fasta + "\n"
#                 ofile.write(fstring)
#     if inplace:
#         output_file = shutil.move(output_file, input_file)
#     return output_file


# def _parse_umis_from_sequences(
#     sequences: Iterable,
#     patterns: Iterable,
#     lengths: Iterable,
#     output_file: str,
#     allowed_mismatches: int = 1,
#     extra_length_for_alignment: int = 25,
#     ignore_strand: bool = False,
#     fmt: str = "fasta",
#     sequence_key: str = "raw_input",
#     id_key: str = "sequence_id",
# ) -> str:
#     """
#     Parses UMIs from a list of sequences.
#     """
#     with open(output_file, "w") as ofile:
#         for s in sequences:
#             if isinstance(s, abutils.Sequence):
#                 if sequence_key in s:
#                     s.sequence = s[sequence_key]
#                 if id_key in s:
#                     s.id = s[id_key]
#             s = abutils.Sequence(s)
#             umi_list = []
#             for p, l in zip(patterns, lengths):
#                 u = UMI(
#                     sequence=s,
#                     pattern=p,
#                     length=l,
#                     extra_length_for_alignment=extra_length_for_alignment,
#                     ignore_strand=ignore_strand,
#                 )
#                 if u.num_mismatches <= allowed_mismatches:
#                     umi_list.append(u.umi)
#             # if we couldn't find a UMI, skip the sequence
#             umi_list = [u for u in umi_list if u is not None]
#             if not umi_list:
#                 continue
#             umi = "+".join(umi_list)
#             s.id = f"{s.id}_{umi}"
#             if fmt.lower() == "fastq":
#                 fstring = s.fastq + "\n"
#             else:
#                 fstring = s.fasta + "\n"
#             ofile.write(fstring)
#     return output_file


# def _get_patterns(name):
#     p = BUILTIN_PATTERNS.get(name, None)
#     if p is not None:
#         return p.get("pattern", None)


# def _get_lengths(name):
#     p = BUILTIN_PATTERNS.get(name, None)
#     if p is not None:
#         return p.get("length", None)


# def _get_allowed_mismatches(name):
#     p = BUILTIN_PATTERNS.get(name, None)
#     if p is not None:
#         return p.get("allowed_mismatches", None)


# # ---------------------------
# #        PATTERNS
# # ---------------------------

# # Takara's Smart-Seq Human BCR kit
# # https://www.takarabio.com/products/next-generation-sequencing/immune-profiling/human-repertoire/smart-seq-human-bcr-with-umis
# SMARTSEQ_HUMAN_BCR_PATTERN = [
#     "[UMI]TCAGCGGGAAGACATT",
#     "[UMI]GGGCGGATGGACTACC",
#     "[UMI]CCGATGGGCGCTTGGT",
#     "[UMI]GGAACACATGCGGAGC",
# ]
# SMARTSEQ_HUMAN_BCR_LENGTH = 12
# SMARTSEQ_HUMAN_BCR_MISMATCH = 2


# BUILTIN_PATTERNS = {
#     "smartseq-human-bcr": {
#         "pattern": SMARTSEQ_HUMAN_BCR_PATTERN,
#         "length": SMARTSEQ_HUMAN_BCR_LENGTH,
#         "allowed_mismatches": SMARTSEQ_HUMAN_BCR_MISMATCH,
#     },
# }
